/*
 * Copyright (c) 1993 Argonaut Software Ltd. All rights reserved.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <dos.h>

#include "brender.h"
#include "dosio.h"
#include "fmt.h"

#define S0 BR_SCALAR(0.0)
#define S1 BR_SCALAR(1.0)

#define MAX_ARG_SIZE 80

/*
 * linked list structure used to process command line args (@file)
 */

static struct list_node {
    	struct list_node *next;
	void *item;
};

void LoadPalette(char *file_name) {
    int i;
    static br_pixelmap *palette = NULL;

    /* Load palette... */

    if (palette!=NULL)
	BR_WARNING0("Overwriting previous palette");

    palette = BrPixelmapLoad(file_name);
    if (palette==NULL)
	BR_ERROR1("Unable to load palette from '%s'",file_name);

    /* ...and install it */

    DOSGfxPaletteSet(palette);
}

void LoadShadeTable(char *file_name) {
    br_pixelmap *shade_table[20];
    int i,n;

    /* Load shade tables... */

    n = BrPixelmapLoadMany(file_name,shade_table,20);
    if (!n)
	BR_ERROR1("Failed to load any shade tables from '%s'",file_name);

    /* ...and add them to registry */

    BrTableAddMany(shade_table,n);
}

void LoadTexture(char *file_name) {
    br_pixelmap *texture[20];
    int i,n;
    char *actual_file_name,*identifier;

    /* Load textures... */

    actual_file_name = strtok(file_name,",");

    n = BrPixelmapLoadMany(file_name,texture,20);
    if (!n)
	BR_ERROR1("Failed to load any textures from '%s'",file_name);

    /* ...and add them to registry */

    for (i = 0; i<n; i++) {
	if ((identifier = strtok(NULL,",")) != NULL)
	    texture[i]->identifier = identifier;
	printf("%s\n",identifier);
	BrMapAdd(texture[i]);
    }
}

void LoadMaterial(char *file_name) {
    br_material *material[20];
    int i,n;

    /* Load materials... */

    n = BrMaterialLoadMany(file_name,material,20);
    if (!n)
	BR_ERROR1("Failed to load any materials from '%s'",file_name);

    /* ...and add them to registry */

    BrMaterialAddMany(material,n);
}

void LoadMaterialScript(char *file_name) {
    br_material *material[20];
    int i,n;

    /* Load materials... */

    n = BrFmtScriptMaterialLoadMany(file_name,material,20);
    if (!n)
	BR_ERROR1("Failed to load any materials from '%s'",file_name);

    /* ...and add them to registry */

    BrMaterialAddMany(material,n);
}

void LoadModel(char *file_name) {
    br_model *model[20];
    int i,n;

    /* Load models... */

    n = BrModelLoadMany(file_name,model,20);
    if (!n)
	BR_ERROR1("Failed to load any models from '%s'",file_name);

    /* ...and add them to registry */

    BrModelAddMany(model,n);
}

br_actor *LoadActor(char *file_name) {
    static br_actor *actor = NULL;

    /* Load actors */

    if (actor!=NULL)
	BR_WARNING0("Overwriting previous actor");

    actor = BrActorLoad(file_name);
    if (actor==NULL)
	BR_ERROR1("Failed to load any actors from '%s'",file_name);

    return actor;
}

/*
 * Internal functions
 */
static char *GetLine(void *fh) {
	char ch,i,*buffer;
	/*
	 * read one line of text from file
	 */
	buffer = BrMemAllocate(MAX_ARG_SIZE,BR_MEMORY_STRING);

	if(BrFileGetLine(buffer,MAX_ARG_SIZE,fh) == 0)
		return NULL;

	return buffer;
}

static void Push(struct list_node **head,void *item)
{
    	struct list_node *new_rec;
	/*
	 * push a pointer to an item onto the head of a list
	 */
	new_rec = BrMemAllocate(sizeof(struct list_node),BR_MEMORY_SCRATCH);
	new_rec->next = *head;
	new_rec->item = item;

	*head = new_rec;
}

static void Pop(struct list_node **head,void **dest)
{
    	struct list_node *curr_node;
	/*
	 * Pop pointer to item from head of list
	 */
	curr_node = *head;
    	*dest = (*head)->item;
	*head = (*head)->next;
	
	BrMemFree(curr_node);
}

static void ProcessArg(int *argc,char ***argv,int *n,int *flag)
{
    	void *fh;
	int open_mode = BR_FS_MODE_TEXT;
	int i,count;
	char *next_line,*sub_arg;
	char **new_argv;
	char *delim = "\n\r ";
	/*
	 * Process command line @file
	 */
	struct list_node *head=NULL;

	fh = BrFileOpenRead( (*argv)[*n]+1,0,NULL,&open_mode);

	if (fh == NULL)
	    BR_ERROR1("Could not open command line file '%s' for reading",(*argv)[*n]);

	for (count = 0;;) {

	    /*
	     * for each line in @file
	     */

	    if ((next_line = GetLine(fh)) == NULL)
		    break;

	    if (*next_line != '#') {
		sub_arg = strtok(next_line,delim);

		while(sub_arg != NULL) {
		    if (*sub_arg=='#') break;

		    /*
		     * store valid entry and get next
		     */	

		    Push(&head,sub_arg);
		    count++;
		    
		    sub_arg = strtok(NULL,delim);
		}

		/*
		 * check for comments in @file
		 */

		if (sub_arg != NULL && *sub_arg == '#')
		    BrMemFree(sub_arg);
	    } else
		BrMemFree(next_line);
	}

	new_argv = malloc(sizeof(*new_argv)*(count+*argc+2));

	/*
	 * generate new argv table of all command line args processed so far
	 */

	for (i = 0; i<*n; i++)
		new_argv[i] = (*argv)[i];

	for (i = *n+count-1; i >= *n; i--)
		Pop(&head,&new_argv[i]);

	for (i = *n+1; i < *argc+1; i++)
		new_argv[i+count-1] = (*argv)[i];

	*argc += count-1;

	if (*flag)
	    BrMemFree(*argv);
	else
	    *flag = 1;

	*argv = new_argv;
	(*n)--;
	
	BrFileClose(fh);
}

char *usage[] = {
    "brview - a BRender model viewer",
    "",
    "-p <palette>\tLoad palette",
    "-h <table>\tLoad shade table",
    "-t <texture>\tLoad texture pixelmap",
    "-s <material>\tLoad material",
    "-m <model>\tLoad model",
    "-a <actor>\tLoad actor",
    "-c \"camera\"\tSelect camera",
    "-l \"light\"\tTurn on light",
    0
};

main(int argc,char *argv[]) {
    br_pixelmap *screen,*back_screen,*depth_buffer;
    br_actor *root,*camera,*light;
    int n,new_cm;

    BR_BANNER("Data Viewer","1994-1995","$Revision: 1.1 $");
#ifndef _MSC_VER
    sleep(2);
#endif

    if (argc<2) {
	char **ptr = usage;

	while (*ptr)
	    fprintf(stderr,"%s\n",*ptr++);

	exit(0);
    }

    BrBegin();
    DOSKeyBegin();
    DOSMouseBegin();

    screen = DOSGfxBegin(NULL);
    back_screen = BrPixelmapMatch(screen,BR_PMMATCH_OFFSCREEN);
    depth_buffer = BrPixelmapMatch(back_screen,BR_PMMATCH_DEPTH_16);
    BrZbBegin(back_screen->type,depth_buffer->type);

    /* Parse arguments */

    /*
     * Process command line to expand @files
     */

    for (n = 1, new_cm = 0; n<argc; n++)
	    if (argv[n][0] == '@')
		    ProcessArg(&argc,&argv,&n,&new_cm);

    while (++argv,--argc) {
	if (!strcmp(argv[0],"-p")) {
	    if (argc<2)
		BR_ERROR0("Specify palette after -p");
	    LoadPalette(argv[1]);
	    ++argv,--argc;
	    continue;
	}
	if (!strcmp(argv[0],"-h")) {
	    if (argc<2)
		BR_ERROR0("Specify shade table after -h");
	    LoadShadeTable(argv[1]);
	    ++argv,--argc;
	    continue;
	}
	if (!strcmp(argv[0],"-t")) {
	    if (argc<2)
		BR_ERROR0("Specify texture file after -t");
	    LoadTexture(argv[1]);
	    ++argv,--argc;
	    continue;
	}
	if (!strcmp(argv[0],"-s")) {
	    if (argc<2)
		BR_ERROR0("Specify material file after -s");
	    LoadMaterial(argv[1]);
	    ++argv,--argc;
	    continue;
	}
	if (!strcmp(argv[0],"-S")) {
	    if (argc<2)
		BR_ERROR0("Specify material script file after -S");
	    LoadMaterialScript(argv[1]);
	    ++argv,--argc;
	    continue;
	}
	if (!strcmp(argv[0],"-m")) {
	    if (argc<2)
		BR_ERROR0("Specify model file after -s");
	    LoadModel(argv[1]);
	    ++argv,--argc;
	    continue;
	}
	if (!strcmp(argv[0],"-a")) {
	    if (argc<2)
		BR_ERROR0("Specify actor file after -a");
	    root = LoadActor(argv[1]);
	    ++argv,--argc;
	    continue;
	}
	if (!strcmp(argv[0],"-c")) {
	    if (argc<2)
		BR_ERROR0("Specify camera after -a");
	    if (root==NULL)
		BR_ERROR0("Can't find camera when no actor loaded");
	    camera = BrActorSearch(root,argv[1]);
	    if (camera==NULL)
		BR_ERROR1("Couldn't find camera '%s'",argv[1]);
	    ++argv,--argc;
	    continue;
	}
	if (!strcmp(argv[0],"-l")) {
	    br_actor *light;

	    if (argc<2)
		BR_ERROR0("Specify light after -a");
	    if (root==NULL)
		BR_ERROR0("Can't find light when no actor loaded");
	    light = BrActorSearch(root,argv[1]);
	    if (light==NULL)
		BR_ERROR1("Couldn't find light '%s'",argv[1]);
	    BrLightEnable(light);
	    ++argv,--argc;
	    continue;
	}
	BR_ERROR1("Erroneous command line option '%s'",argv[0]);
    }

    if (root==NULL)
	BR_ERROR0("No actor loaded");

    if (camera==NULL) {
	camera = BrActorAdd(root,BrActorAllocate(BR_ACTOR_CAMERA,NULL));
	camera->identifier = "camera";
	((br_camera *)camera->type_data)->type = BR_CAMERA_PERSPECTIVE;
	((br_camera *)camera->type_data)->field_of_view = BR_ANGLE_DEG(55.0);
	((br_camera *)camera->type_data)->hither_z = BR_SCALAR(0.01); 
	((br_camera *)camera->type_data)->yon_z = BR_SCALAR(100.0); 
	((br_camera *)camera->type_data)->aspect = BR_SCALAR(1.6); 
	BrMatrix34Translate(&camera->t.t.mat,S0,S0,S1+S1);
    }

    while (!DOSKeyTest(SC_ESC,0,0)) {
	br_matrix34 rotation;
	long x,y;
	unsigned long b;

	BrPixelmapFill(back_screen,0);

	BrPixelmapFill(depth_buffer,0xffffffff);

	BrZbSceneRender(root,camera,back_screen,depth_buffer);

	x = y = 0;
	DOSMouseRead(&x,&y,&b);
	if (x || y) {
	    BrMatrix34RollingBall(&rotation,-x,y,800);
	    BrMatrix34Pre(&camera->t.t.mat,&rotation);
	}
	if (b & BR_MSM_BUTTONL) {
	    BrMatrix34PreTranslate(&camera->t.t.mat,
	        S0,S0,BR_SCALAR(-0.1)
	    );
	}
	if (b & BR_MSM_BUTTONR) {
	    BrMatrix34PreTranslate(&camera->t.t.mat,
	        S0,S0,BR_SCALAR(0.1)
	    );
	}

	BrPixelmapDoubleBuffer(screen,back_screen);
    }

    DOSMouseEnd();
    DOSKeyEnd();
    DOSGfxEnd();
    BrZbEnd();
    BrEnd();
}

